"""Contains exceptions used throughout the program."""


class BadEnvironmentError(Exception):
    """The user has an improperly configured environment."""
    pass


class BadHttpRequestError(Exception):
    """Something bad happened with the HTTP request."""
    pass


class AuthenticationError(BadHttpRequestError):
    """The authentication provided was invalid."""
    pass



"""Container task instance model and manager."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from .base_task_instance import (
    BaseTaskInstance,
    BaseTaskInstanceManager,
)


class ContainerTaskInstance(BaseTaskInstance):
    """Model for container task instances.

    Attributes:
        name (str): The name of the task instance.
        uuid (str): The UUID of the task instance.
        state (str): The state of the task instance.
        user (str): The username of the user who started the task.
        task_queue (int): The ID of the task queue the instance is
            running on.
        task_type (int): The ID of the task type for the instance.
        datetime_created (:class:`datetime.datetime`): The datetime when
            the task instance was created.
        datetime_finished (:class:`datetime.datetime`): The datetime
            when the task instance finished.
        arguments (dict): The arguments the task instance was run with.
    """
    pass


class ContainerTaskInstanceManager(BaseTaskInstanceManager):
    """Manager for container task instances.

    Attributes:
        _client (:py:class:`saltant.client.Client`): An authenticated
            saltant client.
        list_url (str): The URL to list task instances.
        detail_url (str): The URL format to get specific task instances.
        model (:py:class:`saltant.models.resource.Model`): The model of
            the task instance being used.
    """
    list_url = 'containertaskinstances/'
    detail_url = 'containertaskinstances/{uuid}/'
    model = ContainerTaskInstance



"""Contains base classes for models and related classes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from saltant.exceptions import BadHttpRequestError


class Model(object):
    """Base class for representing a model."""
    def __str__(self):
        """String representation of model."""
        raise NotImplementedError


class ModelManager(object):
    """Base class for a model manager.

    Attributes:
        _client (:py:class:`saltant.client.Client`): An authenticated
            saltant client.
        list_url (str): The URL to list models.
        detail_url (str): The URL format to get specific models.
        model (:py:class:`saltant.models.resource.Model`): The model
            being used.
    """
    list_url = "NotImplemented"
    detail_url = "NotImplemented"
    model = Model

    def __init__(self, _client):
        """Save the client so we can make API calls in the manager.

        Args:
            _client (:py:class:`saltant.client.Client`): An
                authenticated saltant client.
        """
        self._client = _client

    def list(self):
        """List instances of models."""
        raise NotImplementedError

    def get(self):
        """Get a specific instance of a model."""
        raise NotImplementedError

    def create(self):
        """Create an instance of a model."""
        raise NotImplementedError

    @classmethod
    def response_data_to_model_instance(cls, response_data):
        """Convert response data to a model.

        Args:
            response_data (dict): The data from the request's response.
        """
        raise NotImplementedError

    @staticmethod
    def validate_request_success(
            request_url,
            status_code,
            expected_status_code,):
        """Validates that a request was successful.

        Args:
            request_url (str): The URL the request was made at.
            status_code (int): The status code of the response.
            expected_status_code (int): The expected status code of the
                response.

        Raises:
            :py:class:`saltant.exceptions.BadHttpRequestError`: The HTTP
                request failed.
        """
        try:
            assert status_code == expected_status_code
        except AssertionError:
            msg = "Request to {} failed with status {}".format(
                request_url,
                status_code,)
            raise BadHttpRequestError(msg)



"""Base class for task instance models."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import json
import dateutil.parser
from saltant.constants import (
    HTTP_200_OK,
    HTTP_201_CREATED,
)
from .resource import Model, ModelManager


class BaseTaskInstance(Model):
    """Base model for a task instance.

    Attributes:
        name (str): The name of the task instance.
        uuid (str): The UUID of the task instance.
        state (str): The state of the task instance.
        user (str): The username of the user who started the task.
        task_queue (int): The ID of the task queue the instance is
            running on.
        task_type (int): The ID of the task type for the instance.
        datetime_created (:class:`datetime.datetime`): The datetime when
            the task instance was created.
        datetime_finished (:class:`datetime.datetime`): The datetime
            when the task instance finished.
        arguments (dict): The arguments the task instance was run with.
    """
    def __init__(
            self,
            uuid,
            state,
            user,
            task_queue,
            task_type,
            datetime_created,
            datetime_finished,
            arguments,
            name="",):
        """Initialize a task instance.

        Args:
            uuid (str): The UUID of the task instance.
            state (str): The state of the task instance.
            user (str): The username of the user who started the task.
            task_queue (int): The ID of the task queue the instance is
                running on.
            task_type (int): The ID of the task type for the instance.
            datetime_created (:class:`datetime.datetime`): The datetime
                when the task instance was created.
            datetime_finished (:class:`datetime.datetime`): The datetime
                when the task instance finished.
            arguments (dict): The arguments the task instance was run
                with.
            name (str, optional): The name of the task instance.
                Defaults to an empty string.
        """
        self.name = name
        self.uuid = uuid
        self.state = state
        self.user = user
        self.task_queue = task_queue
        self.task_type = task_type
        self.datetime_created = datetime_created
        self.datetime_finished = datetime_finished
        self.arguments = arguments

    def __str__(self):
        """String representation of the task instance."""
        return self.uuid


class BaseTaskInstanceManager(ModelManager):
    """Base manager for task instances.

    Attributes:
        _client (:py:class:`saltant.client.Client`): An authenticated
            saltant client.
        list_url (str): The URL to list task instances.
        detail_url (str): The URL format to get specific task instances.
        model (:py:class:`saltant.models.resource.Model`): The model of
            the task instance being used.
    """
    model = BaseTaskInstance

    def get(self, uuid):
        """Get the task instance.

        Args:
            uuid (str): The UUID of the task instance to get.

        Returns:
            :obj:`saltant.models.base_task_instance.BaseTaskInstance`:
                A task instance model instance representing the task
                instance requested.
        """
        # Get the object
        request_url = (
            self._client.base_api_url
            + self.detail_url.format(uuid=uuid))

        response = self._client.session.get(request_url)

        # Validate that the request was successful
        self.validate_request_success(
            request_url=request_url,
            status_code=response.status_code,
            expected_status_code=HTTP_200_OK,)

        # Return a model instance representing the task instance
        return self.response_data_to_model_instance(response.json())

    def create(self,
               task_type_id,
               task_queue_id,
               arguments=None,
               name="",):
        """Create a task instance.

        Args:
            task_type_id (int): The ID of the task type to base the task
                instance on.
            task_queue_id (int): The ID of the task queue to run the job
                on.
            arguments (dict, optional): The arguments to give the task
                type.
            name (str, optional): A non-unique name to give the task
                instance.

        Returns:
            :obj:`saltant.models.base_task_instance.BaseTaskInstance`:
                A task instance model instance representing the task
                instance just created.
        """
        # Make arguments an empty dictionary if None
        if not arguments:
            arguments = {}

        # Create the object
        request_url = self._client.base_api_url + self.list_url
        data_to_post = {
            "name": name,
            "arguments": json.dumps(arguments),
            "task_type": task_type_id,
            "task_queue": task_queue_id,}

        response = self._client.session.post(request_url, data=data_to_post)

        # Validate that the request was successful
        self.validate_request_success(
            request_url=request_url,
            status_code=response.status_code,
            expected_status_code=HTTP_201_CREATED,)

        # Return a model instance representing the task instance
        return self.response_data_to_model_instance(response.json())

    @classmethod
    def response_data_to_model_instance(cls, response_data):
        """Convert response data to a task instance model.

        Args:
            response_data (dict): The data from the request's response.

        Returns:
            :py:obj:`saltant.models.resource.Model`:
                A model instance representing the task instance from the
                reponse data.
        """
        # Coerce datetime strings into datetime objects
        response_data['datetime_created'] = (
            dateutil.parser.parse(response_data['datetime_created']))

        if response_data['datetime_finished']:
            response_data['datetime_finished'] = (
                dateutil.parser.parse(response_data['datetime_finished']))

        # Instantiate a model for the task instance
        return cls.model(**response_data)



"""Contains the saltant API client."""

# TODO(mwiens91): add support for JWT auth tokens

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import functools
import os
import requests
from saltant.constants import (
    DEFAULT_TIMEOUT_SECONDS,
    HTTP_200_OK,
)
from saltant.exceptions import (
    AuthenticationError,
    BadEnvironmentError,
)
from saltant.models.container_task_instance import (
    ContainerTaskInstanceManager,
)
from saltant.models.executable_task_instance import (
    ExecutableTaskInstanceManager,
)


class Client:
    """API client for communicating with a saltant server.

    Example:

        >>> import saltant
        >>> client = saltant.Client(
        ...     base_api_url='https://shahlabjobs.ca/api/',
        ...     auth_token='p0gch4mp101fy451do9uod1s1x9i4a')

    Attributes:
        base_api_url (str): The URL of the saltant API.
        session (:class:`requests.Session`): A session object to make
            requests with.
    """
    def __init__(
            self,
            base_api_url,
            auth_token,
            default_timeout=DEFAULT_TIMEOUT_SECONDS):
        """Initialize the saltant API client.

        Args:
            base_api_url (str): The URL of the saltant API.
            auth_token (str): The registered user's authentication token.
            default_timeout (int, optional): The maximum number
                of seconds to wait for a request to complete. Defaults
                to 90 seconds.
        """
        # The base URL of the saltant API
        self.base_api_url = base_api_url

        # Start a requests session
        self.session = requests.Session()
        self.session.headers.update(
            {'Authorization': 'Token ' + auth_token})

        # Test that we're authorized
        self.test_authentication()

        # Secord the default timeout we want
        self.session.request = functools.partial(
            self.session.request,
            timeout=default_timeout)

        # Add in model managers
        self.container_task_instances = (
            ContainerTaskInstanceManager(_client=self))
        self.executable_task_instances = (
            ExecutableTaskInstanceManager(_client=self))

    def test_authentication(self):
        """Test that the client is authorized.

        This currently assumes that read-only operations require
        authentication, which is the intended authentication protocol
        for saltant servers.

        Raises:
            :py:class:`saltant.exceptions.AuthenticationError`: The
                authentication provided was invalid.
        """
        response = self.session.get(self.base_api_url + 'users/')

        try:
            assert response.status_code == HTTP_200_OK
        except AssertionError:
            raise AuthenticationError('Authentication invalid!')

    @classmethod
    def from_env(cls, default_timeout=DEFAULT_TIMEOUT_SECONDS):
        """Return a client configured from environment variables.

        Essentially copying this:
        https://github.com/docker/docker-py/blob/master/docker/client.py#L43.

        The environment variables looked for are the following:

        .. envvar:: SALTANT_API_URL

            The URL of the saltant API. For example,
            https://shahlabjobs.ca/api/

        .. envvar:: SALTANT_AUTH_TOKEN

            The registered saltant user's authentication token.

        Example:

            >>> import saltant
            >>> client = saltant.from_env()

        Args:
            default_timeout (int, optional): The maximum number of
                seconds to wait for a request to complete. Defaults to
                90 seconds.

        Returns:
            :class:`Client`: A saltant API client object.

        Raises:
            :py:class:`saltant.exceptions.BadEnvironmentError`: The user
                has an incorrectly configured environment.
        """
        # Get variables from environment
        try:
            base_api_url = os.environ['SALTANT_API_URL']
        except KeyError:
            raise BadEnvironmentError("SALTANT_API_URL not defined!")

        try:
            # Try to get an auth token
            auth_token = os.environ['SALTANT_AUTH_TOKEN']
        except KeyError:
            raise BadEnvironmentError("SALTANT_AUTH_TOKEN not defined!")

        # Return the configured client
        return cls(
            base_api_url=base_api_url,
            auth_token=auth_token,
            default_timeout=default_timeout,)


# Allow convenient import access to environment-configured client
from_env = Client.from_env



from setuptools import setup, find_packages
from saltant.version import NAME, DESCRIPTION, VERSION


# Parse readme to include in PyPI page
with open('README.md') as f:
    long_description = f.read()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mwiens91/saltant-py',
    author='Matt Wiens',
    author_email='mwiens91@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    packages=find_packages(),
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    install_requires=[
        'requests',
        'python-dateutil',
    ],
)



"""Contains constants for the program."""

# How many seconds to wait for a request response
DEFAULT_TIMEOUT_SECONDS = 90


# Useful HTTP status codes (nice and verbose :D)
HTTP_200_OK = 200
HTTP_201_CREATED = 201


# Options for the task instance's state field
CREATED = 'created'
PUBLISHED = 'published'
RUNNING = 'running'
SUCCESSFUL = 'successful'
FAILED = 'failed'
TERMINATED = 'terminated'
